
/**
 * tgapi v1.1.0
 * Actual Telegram bot API with Rx-driven updates and full Flow type coverage
 */

'use strict';

Object.defineProperty(exports, '__esModule', { value: true });

function _interopDefault (ex) { return (ex && (typeof ex === 'object') && 'default' in ex) ? ex['default'] : ex; }

var _default = _interopDefault(require('ramda/src/mapObjIndexed'));
var needle = _interopDefault(require('needle'));
var _default$1 = _interopDefault(require('ramda/src/last'));
var Observable = require('rxjs/Observable');
var Subject = require('rxjs/Subject');
require('rxjs/add/observable/from');
require('rxjs/add/operator/map');
require('rxjs/add/operator/startWith');
require('rxjs/add/operator/mergeMap');
require('rxjs/add/operator/publish');
require('rxjs/add/operator/filter');
require('rxjs/add/operator/pluck');
var _default5 = _interopDefault(require('ramda/src/anyPass'));
var _default4 = _interopDefault(require('ramda/src/complement'));
var _default3 = _interopDefault(require('ramda/src/allPass'));
var _default2 = _interopDefault(require('ramda/src/propIs'));
var _default$2 = _interopDefault(require('ramda/src/is'));
var url = require('url');
var http = _interopDefault(require('http'));
var bodyParser = _interopDefault(require('body-parser'));
require('rxjs/add/observable/of');
require('rxjs/add/observable/empty');
require('rxjs/add/observable/merge');
require('rxjs/add/operator/distinctUntilChanged');
require('rxjs/add/operator/distinctUntilKeyChanged');

const prepareBody = _default((value, key) => {
  if (key === 'reply_markup') return JSON.stringify(value);
  if (value instanceof Buffer) return { buffer: value, content_type: 'application/octet-stream' };
  return value;
});

const sendRequest = token => methodName => params => needle('post', `https://api.telegram.org/bot${token}/${methodName}`, params && prepareBody(params), { multipart: true }).then(res => res.body);

var createBotClient = (token => {
  const createMethod = sendRequest(token);

  return {
    get token() {
      return token;
    },

    getMe: createMethod('getMe'),
    getUpdates: createMethod('getUpdates'),
    setWebhook: createMethod('setWebhook'),
    sendMessage: createMethod('sendMessage'),
    forwardMessage: createMethod('forwardMessage'),
    sendPhoto: createMethod('sendPhoto'),
    sendAudio: createMethod('sendAudio'),
    sendDocument: createMethod('sendDocument'),
    sendVideo: createMethod('sendVideo'),
    sendVoice: createMethod('sendVoice'),
    sendVideoNote: createMethod('sendVideoNote'),
    sendMediaGroup: createMethod('sendMediaGroup'),
    sendLocation: createMethod('sendLocation'),
    editMessageLiveLocation: createMethod('editMessageLiveLocation'),
    stopMessageLiveLocation: createMethod('stopMessageLiveLocation'),
    sendVenue: createMethod('sendVenue'),
    sendContact: createMethod('sendContact'),
    sendChatAction: createMethod('sendChatAction'),
    getUserProfilePhotos: createMethod('getUserProfilePhotos'),
    getFile: createMethod('getFile'),
    kickChatMember: createMethod('kickChatMember'),
    unbanChatMember: createMethod('unbanChatMember'),
    restrictChatMember: createMethod('restrictChatMember'),
    promoteChatMember: createMethod('promoteChatMember'),
    exportChatInviteLink: createMethod('exportChatInviteLink'),
    setChatPhoto: createMethod('setChatPhoto'),
    deleteChatPhoto: createMethod('deleteChatPhoto'),
    setChatTitle: createMethod('setChatTitle'),
    setChatDescription: createMethod('setChatDescription'),
    pinChatMessage: createMethod('pinChatMessage'),
    unpinChatMessage: createMethod('unpinChatMessage'),
    leaveChat: createMethod('leaveChat'),
    getChat: createMethod('getChat'),
    getChatAdministrators: createMethod('getChatAdministrators'),
    getChatMembersCount: createMethod('getChatMembersCount'),
    getChatMember: createMethod('getChatMember'),
    setChatStickerSet: createMethod('setChatStickerSet'),
    deleteChatStickerSet: createMethod('deleteChatStickerSet'),
    answerCallbackQuery: createMethod('answerCallbackQuery'),
    editMessageText: createMethod('editMessageText'),
    editMessageCaption: createMethod('editMessageCaption'),
    editMessageReplyMarkup: createMethod('editMessageReplyMarkup'),
    deleteMessage: createMethod('deleteMessage'),
    sendSticker: createMethod('sendSticker'),
    getStickerSet: createMethod('getStickerSet'),
    uploadStickerFile: createMethod('uploadStickerFile'),
    createNewStickerSet: createMethod('createNewStickerSet'),
    addStickerToSet: createMethod('addStickerToSet'),
    setStickerPositionInSet: createMethod('setStickerPositionInSet'),
    deleteStickerFromSet: createMethod('deleteStickerFromSet'),
    answerInlineQuery: createMethod('answerInlineQuery'),
    sendInvoice: createMethod('sendInvoice'),
    answerShippingQuery: createMethod('answerShippingQuery'),
    answerPreCheckoutQuery: createMethod('answerPreCheckoutQuery'),
    sendGame: createMethod('sendGame'),
    setGameScore: createMethod('setGameScore'),
    getGameHighScores: createMethod('getGameHighScores')
  };
});

const getRequesrParams = (timeout, allowedUpdates) => offset => allowedUpdates ? { timeout, offset, allowed_updates: allowedUpdates } : { timeout, offset };

var createUpdateSubscription = ((client, observer$, options = {}) => {
  const getParams = getRequesrParams(Math.max(options.timeout || 1, 1), options.allowedUpdates);

  const subject$ = new Subject.Subject();

  const result$ = subject$.startWith(1).map(getParams).mergeMap(client.getUpdates).publish();

  result$.connect();

  const updates$ = result$.filter(res => res.ok)
  // $FlowFixMe
  .pluck('result').filter(updates => !!updates.length).mergeMap(updates => Observable.Observable.from(updates));

  const offsetSubscription = result$.map(res => {
    if (!res.ok) return 1;
    const lastUpdate = _default$1(res.result);
    return lastUpdate ? lastUpdate.update_id + 1 : 1;
  }).subscribe(subject$);

  const updatesSubscription = updates$.subscribe(observer$);

  return () => {
    offsetSubscription.unsubscribe();
    updatesSubscription.unsubscribe();
  };
});

const isObject = _default$2(Object);

const propIsNumber = _default2(Number);

const propIsObject = propName => _default3([_default2(Object, propName), _default4(_default2(Array, propName))]);

const isValidUpdate = _default3([isObject, propIsNumber('update_id'), _default5([propIsObject('message'), propIsObject('edited_message'), propIsObject('channel_post'), propIsObject('edited_channel_post'), propIsObject('inline_query'), propIsObject('chosen_inline_result'), propIsObject('callback_query'), propIsObject('shipping_query'), propIsObject('pre_checkout_query')])]);

const getHTTPStatusMessage = httpStatusCode => {
  switch (httpStatusCode) {
    case 100:
      return 'Continue';
    case 101:
      return 'Switching Protocols';
    case 102:
      return 'Processing';

    case 200:
      return 'OK';
    case 201:
      return 'Created';
    case 202:
      return 'Accepted';
    case 203:
      return 'Non';
    case 204:
      return 'No Content';
    case 205:
      return 'Reset Content';
    case 206:
      return 'Partial Content';
    case 207:
      return 'Multi';
    case 226:
      return 'IM Used';

    case 300:
      return 'Multiple Choices';
    case 301:
      return 'Moved Permanently';
    case 302:
      return 'Found';
    case 303:
      return 'See Other';
    case 304:
      return 'Not Modified';
    case 305:
      return 'Use Proxy';
    case 306:
      return 'RESERVED in RFC 2616';
    case 307:
      return 'Temporary Redirect';
    case 308:
      return 'Permanent Redirect';

    case 400:
      return 'Bad Request';
    case 401:
      return 'Unauthorized';
    case 402:
      return 'Payment Required';
    case 403:
      return 'Forbidden';
    case 404:
      return 'Not Found';
    case 405:
      return 'Method Not Allowed';
    case 406:
      return 'Not Acceptable';
    case 407:
      return 'Proxy Authentication Required';
    case 408:
      return 'Request Timeout';
    case 409:
      return 'Conflict';
    case 410:
      return 'Gone';
    case 411:
      return 'Length Required';
    case 412:
      return 'Precondition Failed';
    case 413:
      return 'Payload Too Large';
    case 414:
      return 'URI Too Long';
    case 415:
      return 'Unsupported Media Type';
    case 416:
      return 'Range Not Satisfiable';
    case 417:
      return 'Expectation Failed';
    case 418:
      return 'I’m a teapot';
    case 422:
      return 'Unprocessable Entity';
    case 423:
      return 'Locked';
    case 424:
      return 'Failed Dependency';
    case 426:
      return 'Upgrade Required';
    case 428:
      return 'Precondition Required';
    case 429:
      return 'Too Many Requests';
    case 431:
      return 'Request Header Fields Too Large';
    case 434:
      return 'Requested host unavailable';
    case 449:
      return 'Retry With';
    case 451:
      return 'Unavailable For Legal Reasons';

    case 500:
      return 'Internal Server Error';
    case 501:
      return 'Not Implemented';
    case 502:
      return 'Bad Gateway';
    case 503:
      return 'Service Unavailable';
    case 504:
      return 'Gateway Timeout';
    case 505:
      return 'HTTP Version Not Supported';
    case 506:
      return 'Variant Also Negotiates';
    case 507:
      return 'Insufficient Storage';
    case 509:
      return 'Bandwidth Limit Exceeded';
    case 510:
      return 'Not Extended';
    case 511:
      return 'Network Authentication Required';
    case 520:
      return 'Unknown Error';
    case 521:
      return 'Web Server Is Down';
    case 522:
      return 'Connection Timed Out';
    case 523:
      return 'Origin Is Unreachable';
    case 524:
      return 'A Timeout Occurred';
    case 525:
      return 'SSL Handshake Failed';
    case 526:
      return 'Invalid SSL Certificate';

    default:
      {

        return 'Unknown HTTP error';
      }
  }
};

class HTTPError extends Error {

  static fromCatch(error) {
    if (error instanceof HTTPError) return error;

    return new HTTPError(500, error);
  }

  constructor(httpStatusCode, error) {
    if (error instanceof Error) {
      super(error.message);

      this.name = error.name;
      this.fileName = error.fileName;
      this.lineNumber = error.lineNumber;
      this.columnNumber = error.columnNumber;
      this.stack = error.stack;
    } else {
      super(error);
    }

    this.name = 'HTTPError';
    this.httpStatusCode = httpStatusCode;
    this.httpStatusMessage = getHTTPStatusMessage(httpStatusCode);
  }
}

/* :: interface HTTPServer extends Server {} */

const createErrorHandler = (bot, res) => err => {
  const error = HTTPError.fromCatch(err);

  res.statusCode = error.httpStatusCode;
  res.statusMessage = error.httpStatusMessage;

  res.end();

  bot.error && bot.error(error);
};

var setWebhookToServer = ((bot, path = '') => server => {
  const url$$1 = url.resolve('/', path);

  server.on('request', (req, res) => {
    const handleError = createErrorHandler(bot, res);

    if (req.method !== 'POST') {
      handleError(new HTTPError(req.url === url$$1 ? 405 : 501, `Invalid ${req.method} request to ${req.url}`));
    } else if (req.url !== url$$1) {
      handleError(new HTTPError(404, `Invalid ${req.method} request to ${req.url}`));
    } else {
      let data = '';

      req.on('data', newData => data += newData || '');

      req.on('end', () => {
        try {
          const update = JSON.parse(data);

          if (!isValidUpdate(update)) throw new HTTPError(400, 'Bad request');

          bot.next(update);
          res.end();
        } catch (e) {
          handleError(e);
        }
      });
    }
  });
});

var createWebhookServer = ((bot$, path) => {
  const server = http.createServer();

  setWebhookToServer(bot$, path)(server);

  return server;
});

var createMiddleware = (observer => {
  const parserMiddleware = bodyParser.json();

  const botMiddlware = ({ body }, res) => {
    if (isValidUpdate(body)) {
      observer.next(body);
      res.end();
      return;
    }

    if (observer.error) observer.error(new HTTPError(400, 'Invalid update'));
    res.status(400).send('Bad Request');
  };

  return (req, res) => typeof req.body === 'object' ? botMiddlware(req, res) : parserMiddleware(req, res, () => botMiddlware(req, res));
});

const formatUpdate = ({
  /* eslint-disable camelcase */
  update_id,
  message,
  edited_message,
  channel_post,
  edited_channel_post,
  inline_query,
  chosen_inline_result,
  callback_query,
  shipping_query,
  pre_checkout_query
}) => {
  if (message) return { update_id, message, type: 'message' };
  if (edited_message) return { update_id, edited_message, type: 'edited_message' };
  if (channel_post) return { update_id, channel_post, type: 'channel_post' };
  if (edited_channel_post) return { update_id, edited_channel_post, type: 'edited_channel_post' };
  if (inline_query) return { update_id, inline_query, type: 'inline_query' };
  if (chosen_inline_result) return { update_id, chosen_inline_result, type: 'chosen_inline_result' };
  if (callback_query) return { update_id, callback_query, type: 'callback_query' };
  if (shipping_query) return { update_id, shipping_query, type: 'shipping_query' };
  if (pre_checkout_query) return { update_id, pre_checkout_query, type: 'pre_checkout_query'
    /* eslint-enable camelcase */
  };
};

var createUpdateObserver = (() => {
  const rawUpdate$ = new Subject.Subject();

  const update$ = rawUpdate$.map(formatUpdate).mergeMap(update => update ? Observable.Observable.of(update) : Observable.Observable.empty()).publish();

  const message$ = update$.mergeMap(update => update.type === 'message' ? Observable.Observable.of(update) : Observable.Observable.empty()).publish();

  const editedMessage$ = update$.mergeMap(update => update.type === 'edited_message' ? Observable.Observable.of(update) : Observable.Observable.empty()).publish();

  const channelPost$ = update$.mergeMap(update => update.type === 'channel_post' ? Observable.Observable.of(update) : Observable.Observable.empty()).publish();

  const editedChannelPost$ = update$.mergeMap(update => update.type === 'edited_channel_post' ? Observable.Observable.of(update) : Observable.Observable.empty()).publish();

  const inlineQuery$ = update$.mergeMap(update => update.type === 'inline_query' ? Observable.Observable.of(update) : Observable.Observable.empty()).publish();

  const chosenInlineResult$ = update$.mergeMap(update => update.type === 'chosen_inline_result' ? Observable.Observable.of(update) : Observable.Observable.empty()).publish();

  const callbackQuery$ = update$.mergeMap(update => update.type === 'callback_query' ? Observable.Observable.of(update) : Observable.Observable.empty()).publish();

  const shippingQuery$ = update$.mergeMap(update => update.type === 'shipping_query' ? Observable.Observable.of(update) : Observable.Observable.empty()).publish();

  const preCheckoutQuery$ = update$.mergeMap(update => update.type === 'pre_checkout_query' ? Observable.Observable.of(update) : Observable.Observable.empty()).publish();

  update$.connect();
  message$.connect();
  editedMessage$.connect();
  channelPost$.connect();
  editedChannelPost$.connect();
  inlineQuery$.connect();
  chosenInlineResult$.connect();
  callbackQuery$.connect();
  shippingQuery$.connect();
  preCheckoutQuery$.connect();

  return {
    next: update => rawUpdate$.next(update),
    error: _error => rawUpdate$.error(_error),
    complete: () => rawUpdate$.complete(),

    update$,
    message$,
    editedMessage$,
    channelPost$,
    editedChannelPost$,
    inlineQuery$,
    chosenInlineResult$,
    callbackQuery$,
    shippingQuery$,
    preCheckoutQuery$
  };
});

exports.createBotClient = createBotClient;
exports.createUpdateSubscription = createUpdateSubscription;
exports.createWebhookServer = createWebhookServer;
exports.createMiddleware = createMiddleware;
exports.createUpdateObserver = createUpdateObserver;
